#include "Application.h"
#include <stdarg.h>        // For argument processing
#include <uart_regs.h>

extern const uint16_t EpSize[];

// Declare external functions
extern void CheckStatus(char* StringPtr, CyU3PReturnStatus_t Status);

CyBool_t glIsUartActive = CyFalse;      // Set true once USB <-> UART communication is established

CyU3PDmaChannel Usb2Uart_Handle;        // Handler to USB to UART DMA
CyU3PDmaChannel Uart2Usb_Handle;        // Handler to UART to USB DMA
volatile uint16_t glPktsPending = 0;    // Number of packets that have been committed since last check
uint32_t regValueEn = 0, regValueDs = 0;


CyU3PReturnStatus_t ConfigureUart(void)
{
    CyU3PReturnStatus_t Status = CY_U3P_SUCCESS;
    CyU3PUartConfig_t uartConfig;

    // Start the UART driver
    Status = CyU3PUartInit();
    CheckStatus("CyU3PUartInit", Status);

    // Configure my UART link, as fast as possible for interfacing with FPGA
    CyU3PMemSet ((uint8_t *)&uartConfig, 0, sizeof (uartConfig));
    uartConfig.baudRate = CY_U3P_UART_BAUDRATE_4M608K;
    uartConfig.stopBit  = CY_U3P_UART_ONE_STOP_BIT;
    uartConfig.txEnable = CyTrue;
    uartConfig.rxEnable = CyTrue;
    uartConfig.flowCtrl = CyFalse;
    uartConfig.isDma    = CyTrue;

    /* Set the UART configuration; no callback  */
    Status = CyU3PUartSetConfig(&uartConfig, NULL);
    CheckStatus("CyU3PUartSetConfig", Status);

    Status = CyU3PUartTxSetBlockXfer(0xFFFFFFFF);                   // Send as much data as I need to
    CheckStatus("CyU3PUartTxSetBlockXfer", Status);

    // Status = CyU3PUartRxSetBlockXfer(0xFFFFFFFF);
    // CheckStatus("CyU3PUartRxSetBlockXfer", Status);

    /* UART Config Value for Enabling Rx Block */
    regValueEn = UART->lpp_uart_config;

    /* UART Config Value for Disabling the Rx Block  */
    regValueDs = UART->lpp_uart_config & (~(CY_U3P_LPP_UART_RTS | CY_U3P_LPP_UART_RX_ENABLE));


    return Status;
}

/* Callback function for the DMA event notification. */
void UartRxDMACallback(
        CyU3PDmaChannel   *chHandle, /* Handle to the DMA channel. */
        CyU3PDmaCbType_t   type,     /* Callback type.             */
        CyU3PDmaCBInput_t *input)    /* Callback status.           */
{
    CyU3PReturnStatus_t Status = CY_U3P_SUCCESS;
    if (type == CY_U3P_DMA_CB_PROD_EVENT)
    {
        DebugPrint(4, "\r\nUartRxDMACallback call");
        Status = CyU3PDmaChannelCommitBuffer (&Uart2Usb_Handle, input->buffer_p.count, 0);
        CheckStatus("CyU3PDmaChannelCommitBuffer()", Status);
        glPktsPending++;
    }
}
//}

CyU3PReturnStatus_t ConfigureUartEP(void)
{
    CyU3PReturnStatus_t Status = CY_U3P_SUCCESS;
    CyU3PEpConfig_t epCfg;

    // Based on the Bus Speed configure the endpoint packet size
    CyU3PUSBSpeed_t usbSpeed = CyU3PUsbGetSpeed();

    // Configure FPGA <-> UART Endpoints
    CyU3PMemSet((uint8_t *)&epCfg, 0, sizeof(epCfg));
    epCfg.enable = CyTrue;
    epCfg.epType = CY_U3P_USB_EP_BULK;
    epCfg.burstLen = (usbSpeed == CY_U3P_SUPER_SPEED) ? (ENDPOINT_BURST_LENGTH) : 1;
    epCfg.pcktSize = EpSize[usbSpeed];

    // Setup and flush the FPGA_CONFIG endpoint
    Status = CyU3PSetEpConfig(FPGA_CONFIG_ENDPOINT, &epCfg);
    CheckStatus("Setup FPGA_CONFIG_ENDPOINT", Status);

    Status = CyU3PUsbFlushEp(FPGA_CONFIG_ENDPOINT);
    CheckStatus("Flush FPGA_CONFIG_ENDPOINT", Status);

    // Setup and flush the FPGA_STATUS endpoint
    Status = CyU3PSetEpConfig(FPGA_STATUS_ENDPOINT, &epCfg);
    CheckStatus("Setup FPGA_STATUS_ENDPOINT", Status);

    Status = CyU3PUsbFlushEp(FPGA_STATUS_ENDPOINT);
    CheckStatus("Flush FPGA_STATUS_ENDPOINT", Status);

    return Status;
}

CyU3PReturnStatus_t ConfigureUartDma(void) {
    CyU3PReturnStatus_t Status = CY_U3P_SUCCESS;
    CyU3PDmaChannelConfig_t dmaConfig;

    // Create a AUTO channel for the USB to UART transfer
    CyU3PMemSet((uint8_t *)&dmaConfig, 0, sizeof(dmaConfig));
    dmaConfig.size            = 1024; // Maximal number of bytes allowed per DMA buffer.
                                     // A larger message would stuck the EP
    dmaConfig.count           = 8;  // Store up to 32 words in DMA buffers
    dmaConfig.prodSckId       = FPGA_CONFIG_ENDPOINT_SOCKET;
    dmaConfig.consSckId       = UART_CONSUMER_SOCKET;
    dmaConfig.dmaMode         = CY_U3P_DMA_MODE_BYTE;
    dmaConfig.notification    = 0;
    dmaConfig.cb              = NULL;
    // dmaConfig.prodHeader      = 0;
    // dmaConfig.prodFooter      = 0;
    // dmaConfig.consHeader      = 0;
    // dmaConfig.prodAvailCount  = 0;

    Status = CyU3PDmaChannelCreate(&Usb2Uart_Handle, CY_U3P_DMA_TYPE_AUTO, &dmaConfig);
    CheckStatus("Usb2Uart DmaChannelCreate", Status);

    // Start the DMA Channel with transfer size to Infinite
    Status = CyU3PDmaChannelSetXfer(&Usb2Uart_Handle, 0);
    CheckStatus("Usb2Uart DmaChannelStart", Status);

    // Create a AUTO SIGNAL channel for the UART to USB transfer
    CyU3PMemSet((uint8_t *)&dmaConfig, 0, sizeof(dmaConfig));
    dmaConfig.size          = 32;   // Minimal value allowed by DMA.
                                    // Any UART message longer than this will be split in 16 bytes "packets"
    dmaConfig.count         = 8;   // Up to 32 words can be stored in DMA buffers
    dmaConfig.prodSckId     = UART_PRODUCER_SOCKET;
    dmaConfig.consSckId     = FPGA_STATUS_ENDPOINT_SOCKET;
    dmaConfig.dmaMode       = CY_U3P_DMA_MODE_BYTE;
    dmaConfig.notification  = CY_U3P_DMA_CB_PROD_EVENT;
    dmaConfig.cb            = UartRxDMACallback;
    Status = CyU3PDmaChannelCreate(&Uart2Usb_Handle, CY_U3P_DMA_TYPE_AUTO_SIGNAL, &dmaConfig);
    CheckStatus("Uart2Usb DmaChannelCreate", Status);

    // Start the DMA Channel with transfer size to Infinite
    Status = CyU3PDmaChannelSetXfer(&Uart2Usb_Handle, 0);
    CheckStatus("Uart2Usb DmaChannelStart", Status);

    return Status;
}

CyU3PReturnStatus_t StartUart(void)
{
    CyU3PReturnStatus_t Status = CY_U3P_SUCCESS;

    // Configure UART for interfacing with FPGA
    Status = ConfigureUart();
    CheckStatus("ConfigureUart()", Status);

    // Configure USB endpoints linked to UART
    Status = ConfigureUartEP();
    CheckStatus("ConfigureUartEP()", Status);

    // Create DMAs between EP and UART
    Status = ConfigureUartDma();
    CheckStatus("ConfigureUartDma()", Status);

    // Mark USB <-> UART communication as active
    glIsUartActive = CyTrue;

    return Status;
}

CyU3PReturnStatus_t StopUart(void)
{
    CyU3PReturnStatus_t Status = CY_U3P_SUCCESS;
    CyU3PEpConfig_t epConfig;

    // Mark USB <-> UART communication as inactive
    glIsUartActive = CyFalse;

    // Destroy UART Related DMAs
    Status = CyU3PDmaChannelDestroy(&Usb2Uart_Handle);
    CheckStatus("Usb2Uart DmaChannelDestroy", Status);
    Status = CyU3PDmaChannelDestroy(&Uart2Usb_Handle);
    CheckStatus("Uart2Usb DmaChannelDestroy", Status);

    // Flush and disable Endpoints
    CyU3PUsbFlushEp(FPGA_CONFIG_ENDPOINT);
    CheckStatus("Flush FPGA_CONFIG_ENDPOINT", Status);
    CyU3PUsbFlushEp(FPGA_STATUS_ENDPOINT);
    CheckStatus("Flush FPGA_STATUS_ENDPOINT", Status);
    CyU3PMemSet((uint8_t *)&epConfig, 0, sizeof(&epConfig));
    Status = CyU3PSetEpConfig(FPGA_CONFIG_ENDPOINT, &epConfig);
    CheckStatus("Disable FPGA_CONFIG_ENDPOINT", Status);
    Status = CyU3PSetEpConfig(FPGA_STATUS_ENDPOINT, &epConfig);
    CheckStatus("Disable FPGA_STATUS_ENDPOINT", Status);

    // Stop UART
    Status = CyU3PUartDeInit();
    CheckStatus("Disable UART", Status);

    return Status;
}

CyU3PReturnStatus_t FlushUARTRxDMA(void)
{
    CyU3PReturnStatus_t Status = CY_U3P_SUCCESS;
    if (glPktsPending == 0)
    {
        /* Disable UART Receiver Block */
        UART->lpp_uart_config = regValueDs;

        Status = CyU3PDmaChannelSetWrapUp (&Uart2Usb_Handle);

        /* Enable UART Receiver Block */
        UART->lpp_uart_config = regValueEn;
    }

    glPktsPending = 0;
    return Status;
}

